// src/hooks/graph-populate.hook.ts
import _isEmpty2 from "lodash/isEmpty.js";

// src/hooks/shallow-populate.hook.ts
import _get2 from "lodash/get.js";
import _set2 from "lodash/set.js";

// src/utils/shallow-populate.utils.ts
import _get from "lodash/get.js";
import _isEmpty from "lodash/isEmpty.js";
import _isEqual from "lodash/isEqual.js";
import _isFunction from "lodash/isFunction.js";
import _merge from "lodash/merge.js";
import _set from "lodash/set.js";
import _uniqBy from "lodash/uniqBy.js";

// src/utils/to-array.ts
var toArray = (value) => {
  if (value == null) {
    return [];
  }
  return Array.isArray(value) ? value : [value];
};

// src/utils/shallow-populate.utils.ts
var requiredIncludeAttrs = ["service", "nameAs", "asArray", "params"];
var isDynamicParams = (params) => {
  if (!params) return false;
  if (Array.isArray(params)) {
    return params.some((p) => isDynamicParams(p));
  } else {
    return !_isEmpty(params) || _isFunction(params);
  }
};
var shouldCatchOnError = (options, include) => {
  if (include.catchOnError !== void 0) return !!include.catchOnError;
  if (options.catchOnError !== void 0) return !!options.catchOnError;
  return false;
};
var assertIncludes = (includes) => {
  for (const include of includes) {
    if (!("asArray" in include)) {
      include.asArray = true;
    }
    if (!("params" in include)) {
      include.params = {};
    }
    if (!("requestPerItem" in include)) {
      include.requestPerItem = !("keyHere" in include) && !("keyThere" in include);
    }
    const isDynamic = isDynamicParams(include.params);
    const requiredAttrs = isDynamic ? requiredIncludeAttrs : [...requiredIncludeAttrs, "keyHere", "keyThere"];
    requiredAttrs.forEach((attr) => {
      if (!(attr in include)) {
        throw new Error(
          "shallowPopulate hook: Every `include` must contain `service`, `nameAs` and (`keyHere` and `keyThere`) or `params` properties"
        );
      }
    });
    if (isDynamic && (!("keyHere" in include) && "keyThere" in include || "keyHere" in include && !("keyThere" in include))) {
      throw new Error(
        "shallowPopulate hook: Every `include` with attribute `keyHere` or `keyThere` also needs the other attribute defined"
      );
    }
    if (include.requestPerItem && ("keyHere" in include || "keyThere" in include)) {
      throw new Error(
        "shallowPopulate hook: The attributes `keyHere` and `keyThere` are useless when you set `requestPerItem: true`. You should remove these properties"
      );
    }
  }
  const uniqueNameAs = _uniqBy(includes, "nameAs");
  if (uniqueNameAs.length !== includes.length) {
    throw new Error(
      "shallowPopulate hook: Every `\xECnclude` must have a unique `nameAs` property"
    );
  }
};
var chainedParams = async (paramsArr, context, target, options = {}) => {
  if (!Array.isArray(paramsArr)) paramsArr = [paramsArr];
  const { thisKey, skipWhenUndefined } = options;
  const resultingParams = {};
  for (let i = 0, n = paramsArr.length; i < n; i++) {
    let params = paramsArr[i];
    if (_isFunction(params)) {
      params = thisKey == null ? params(resultingParams, context, target) : params.call(thisKey, resultingParams, context, target);
      params = await Promise.resolve(params);
    }
    if (!params && skipWhenUndefined) return void 0;
    if (params !== resultingParams) _merge(resultingParams, params);
  }
  return resultingParams;
};
async function makeCumulatedRequest(app, { include, ids }, context) {
  const { keyThere } = include;
  let params = {
    query: {
      [keyThere]: ids.length === 1 ? ids[0] : { $in: ids }
    },
    paginate: false
  };
  const service = app.service(include.service);
  const target = {
    path: include.service,
    service
  };
  params = await chainedParams(
    [params, ...toArray(include.params)],
    context,
    target
  );
  const { $skip, $limit, ...query } = params.query ? { ...params.query } : {};
  if (query.$select && !query.$select.includes(keyThere)) {
    query.$select = [...query.$select, keyThere];
  }
  const response = await service.find({ ...params, query });
  return {
    include,
    params,
    response
  };
}
async function makeRequestPerItem(item, app, include, context) {
  const { nameAs, asArray } = include;
  const paramsFromInclude = toArray(include.params);
  const paramsOptions = {
    thisKey: item,
    skipWhenUndefined: true
  };
  const service = app.service(include.service);
  const target = {
    path: include.service,
    service
  };
  const params = await chainedParams(
    [{ paginate: false }, ...paramsFromInclude],
    context,
    target,
    paramsOptions
  );
  if (!params) {
    _set(item, nameAs, noRelation(include));
    return;
  }
  const relatedItems = await service.find(params);
  if (asArray) {
    _set(item, nameAs, relatedItems);
  } else {
    const relatedItem = relatedItems.length > 0 ? relatedItems[0] : null;
    _set(item, nameAs, relatedItem);
  }
}
function setItems(data, include, params, relatedItems) {
  const { nameAs, keyThere, asArray } = include;
  data.forEach((item) => {
    const keyHere = _get(item, include.keyHere);
    if (keyHere !== void 0) {
      if (Array.isArray(keyHere)) {
        if (!asArray) {
          const items = getRelatedItems(
            keyHere[0],
            relatedItems,
            include,
            params
          );
          if (items !== void 0) {
            _set(item, nameAs, items);
          }
        } else {
          _set(
            item,
            nameAs,
            getRelatedItems(keyHere, relatedItems, include, params)
          );
        }
      } else {
        const items = getRelatedItems(keyHere, relatedItems, include, params);
        if (items !== void 0) {
          _set(item, nameAs, items);
        }
      }
    }
  });
  if (params.query?.$select && !params.query.$select.includes(keyThere)) {
    relatedItems.forEach((item) => {
      delete item[keyThere];
    });
  }
}
function getRelatedItems(ids, relatedItems, include, params) {
  const { keyThere } = include;
  const skip = params?.query?.$skip ?? 0;
  const limit = params?.query?.$limit ?? Math.max;
  ids = toArray(ids);
  let skipped = 0;
  let itemOrItems = noRelation(include);
  let isDone = false;
  for (let i = 0, n = relatedItems.length; i < n; i++) {
    if (isDone) break;
    const currentItem = relatedItems[i];
    for (let j = 0, m = ids.length; j < m; j++) {
      const id = ids[j];
      let currentId;
      if (keyThere.includes(".") && Array.isArray(currentItem[keyThere.slice(0, keyThere.indexOf("."))])) {
        const arrayName = keyThere.split(".")[0];
        const nestedProp = keyThere.slice(keyThere.indexOf(".") + 1);
        currentId = currentItem[arrayName].map((nestedItem) => {
          const keyThereVal = _get(nestedItem, nestedProp);
          return keyThereVal;
        });
      } else {
        const keyThereVal = _get(currentItem, keyThere);
        currentId = keyThereVal;
      }
      if (include.asArray) {
        const items = itemOrItems;
        if (Array.isArray(currentId) && currentId.includes(id) || _isEqual(currentId, id)) {
          if (skipped < skip) {
            skipped++;
            continue;
          }
          items.push(currentItem);
          if (items.length >= limit) {
            isDone = true;
            break;
          }
        }
      } else {
        if (_isEqual(currentId, id)) {
          if (skipped < skip) {
            skipped++;
            continue;
          }
          itemOrItems = currentItem;
          isDone = true;
          break;
        }
      }
    }
  }
  return itemOrItems;
}
var noRelation = (include) => {
  return include.asArray ? [] : null;
};

// src/hooks/shallow-populate.hook.ts
function shallowPopulate(opts) {
  const options = {
    catchOnError: false,
    ...opts
  };
  const includes = toArray(options.include);
  if (!includes.length) {
    throw new Error(
      "shallowPopulate hook: You must provide one or more relationships in the `include` option."
    );
  }
  assertIncludes(includes);
  const cumulatedIncludes = includes.filter(
    (include) => !include.requestPerItem
  );
  const includesPerItem = includes.filter(
    (include) => include.requestPerItem
  );
  return async (context) => {
    const { app, type } = context;
    let data = type === "before" ? context.data : context.method === "find" ? context.result.data || context.result : context.result;
    data = toArray(data);
    if (!data.length) {
      return context;
    }
    const includesAndIds = [];
    for (const include of cumulatedIncludes) {
      let result = void 0;
      for (const item of data) {
        const id = _get2(item, include.keyHere);
        if (id == null) {
          _set2(item, include.nameAs, noRelation(include));
          continue;
        }
        if (!result) {
          result = { include, ids: [] };
        }
        const ids = toArray(id);
        for (const id2 of ids) {
          if (!result.ids.includes(id2)) {
            result.ids.push(id2);
          }
        }
      }
      if (result) {
        includesAndIds.push(result);
      }
    }
    const promises = [];
    for (const keyHere in includesAndIds) {
      const includeAndIds = includesAndIds[keyHere];
      if (!includeAndIds.ids.length) continue;
      promises.push(
        new Promise((resolve, reject) => {
          makeCumulatedRequest(app, includeAndIds, context).then(resolve).catch((err) => {
            if (!shouldCatchOnError(options, includeAndIds.include)) {
              reject(err);
              return;
            }
            resolve({ include: includeAndIds.include });
          });
        })
      );
    }
    const cumulatedResults = await Promise.all(promises);
    cumulatedResults.forEach((result) => {
      if (!result) return;
      const { include } = result;
      if (!result.response) {
        data.forEach((item) => {
          _set2(item, include.nameAs, noRelation(include));
        });
        return;
      }
      const { params, response } = result;
      setItems(data, include, params, response);
    });
    const promisesPerIncludeAndItem = [];
    includesPerItem.forEach((include) => {
      const promisesPerItem = data.map(async (item) => {
        try {
          await makeRequestPerItem(item, app, include, context);
        } catch (err) {
          if (!shouldCatchOnError(options, include)) throw err;
          _set2(item, include.nameAs, noRelation(include));
        }
      });
      promisesPerIncludeAndItem.push(...promisesPerItem);
    });
    await Promise.all(promisesPerIncludeAndItem);
    return context;
  };
}

// src/hooks/graph-populate.hook.ts
var FILTERS = ["$limit", "$select", "$skip", "$sort"];
function graphPopulate(options) {
  if (!options.populates) {
    throw new Error(
      "options.populates must be provided to the feathers-graph-populate hook"
    );
  }
  const { populates } = options;
  return async (context) => {
    const populateQuery = context.params?.$populateParams?.query;
    if (!populateQuery) return context;
    const { app } = context;
    const graphPopulateApp = app.graphPopulate;
    const keys = Object.keys(populateQuery);
    const currentPopulates = keys.reduce((currentPopulates2, key) => {
      if (!populates[key]) return currentPopulates2;
      const currentQuery = { ...populateQuery[key] };
      const populate2 = populates[key];
      const service = app.service(populate2.service);
      let params = [];
      if (populate2.params) {
        if (Array.isArray(populate2.params)) {
          params.push(...populate2.params);
        } else {
          params.push(populate2.params);
        }
      }
      if (!_isEmpty2(currentQuery)) {
        const customKeysForQuery = service.options?.graphPopulate?.whitelist;
        const extractKeys = [...FILTERS];
        if (customKeysForQuery) {
          extractKeys.push(...customKeysForQuery);
        }
        const paramsToAdd = Object.keys(currentQuery).reduce(
          (paramsToAdd2, key2) => {
            if (!extractKeys.includes(key2)) return paramsToAdd2;
            const { query } = paramsToAdd2;
            query[key2] = currentQuery[key2];
            delete currentQuery[key2];
            return paramsToAdd2;
          },
          { query: {} }
        );
        params.push(paramsToAdd);
      }
      if (!_isEmpty2(currentQuery)) {
        params.push({
          $populateParams: {
            query: currentQuery
          }
        });
      }
      if (graphPopulateApp) {
        params = graphPopulateApp.withAppParams(
          params,
          context.method,
          service
        );
      }
      currentPopulates2.push({
        ...populate2,
        params
      });
      return currentPopulates2;
    }, []);
    if (!currentPopulates?.length) {
      return context;
    }
    const shallowPopulate2 = shallowPopulate({ include: currentPopulates });
    const populatedContext = await shallowPopulate2(context);
    return populatedContext;
  };
}

// src/utils/get-query.ts
function getQuery(options) {
  const { context, namedQueries } = options;
  let query = context.params?.$populateParams?.query;
  const allowByHook = options.allowUnnamedQueryForExternal;
  const allowByApp = context.app.graphPopulate?.options?.allowUnnamedQueryForExternal;
  if (query && context.params.provider && !(allowByHook ?? allowByApp)) {
    delete context.params.$populateParams.query;
    query = void 0;
  }
  if (!query) {
    if (!namedQueries) {
      return void 0;
    }
    const name = context.params?.$populateParams?.name || options.defaultQueryName;
    if (!name) {
      return void 0;
    }
    query = namedQueries[name]?.query || namedQueries[name];
  }
  return query;
}

// src/hooks/populate.hook.ts
function populate(options) {
  const {
    namedQueries,
    defaultQueryName,
    populates,
    allowUnnamedQueryForExternal
  } = options;
  const graphPopulate2 = graphPopulate({ populates });
  return async (context) => {
    if (!context.params.$populateParams && !defaultQueryName) {
      return context;
    }
    const query = getQuery({
      context,
      namedQueries,
      defaultQueryName,
      allowUnnamedQueryForExternal
    });
    if (!query) return context;
    if (!context.params.$populateParams) context.params.$populateParams = {};
    context.params.$populateParams.query = query;
    return graphPopulate2(context);
  };
}

// src/utils/util.populate.ts
async function populateUtil(records, options) {
  const { app, params, populates } = options;
  if (!app) {
    throw new Error(
      "The app object must be provided in the populateUtil options."
    );
  }
  if (!("$populateParams" in params)) {
    return records;
  }
  const $populateParams = params.$populateParams;
  const populateQuery = $populateParams.query;
  if (!populates || !populateQuery || !Object.keys(populateQuery).length) {
    return records;
  }
  const miniContext = {
    app,
    method: "find",
    type: "after",
    result: records,
    params
  };
  const deepPopulate = graphPopulate({ populates });
  const populated = await deepPopulate(miniContext);
  return populated.result;
}

// src/hooks/params-for-server.hook.ts
function paramsForServer(...whitelist) {
  return (context) => {
    const params = JSON.parse(JSON.stringify(context.params));
    params.query = params.query || {};
    params.query._$client = params.query._$client || {};
    Object.keys(params).forEach((key) => {
      if (key !== "query") {
        if (whitelist.includes(key)) {
          params.query._$client[`_${key}`] = params[key];
          delete context.params[key];
        }
      }
    });
    context.params = params;
  };
}

// src/hooks/params-from-client.hook.ts
function paramsFromClient(...whitelist) {
  return (context) => {
    const params = context.params;
    if (params && params.query && params.query._$client && typeof params.query._$client === "object") {
      const client = params.query._$client;
      whitelist.forEach((key) => {
        if (`_${key}` in client) {
          params[key] = client[`_${key}`];
        }
      });
      params.query = Object.assign({}, params.query);
      delete params.query._$client;
    }
    return context;
  };
}

// src/utils/define-populates.ts
var definePopulates = (populates) => {
  return populates;
};

// src/app/hooks.commons.ts
import { _ } from "@feathersjs/commons";
import _get3 from "lodash/get.js";
var { each } = _;
function convertHookData(obj) {
  let hook = {};
  if (Array.isArray(obj)) {
    hook = { all: obj };
  } else if (typeof obj !== "object") {
    hook = { all: [obj] };
  } else {
    each(obj, function(value, key) {
      hook[key] = !Array.isArray(value) ? [value] : value;
    });
  }
  return hook;
}
function getHooks(app, service, type, method, appLast = false) {
  const appHooks = _get3(app, ["__hooks", type, method]) || [];
  const serviceHooks = _get3(service, ["__hooks", type, method]) || [];
  return appLast ? [...serviceHooks, ...appHooks] : [...appHooks, ...serviceHooks];
}
function enableHooks(obj, methods, types) {
  if (typeof obj.hooks === "function") {
    return obj;
  }
  const hookData = {};
  types.forEach((type) => {
    hookData[type] = {};
  });
  Object.defineProperty(obj, "__hooks", {
    configurable: true,
    value: hookData,
    writable: true
  });
  return Object.assign(obj, {
    hooks(allHooks) {
      each(
        allHooks,
        (current, type) => {
          if (!this.__hooks[type]) {
            throw new Error(`'${type}' is not a valid hook type`);
          }
          const hooks = convertHookData(current);
          methods.forEach((method) => {
            const currentHooks = this.__hooks[type][method] || (this.__hooks[type][method] = []);
            if (hooks.all) {
              currentHooks.push(...hooks.all);
            }
            if (hooks[method]) {
              currentHooks.push(...hooks[method]);
            }
          });
        }
      );
      return this;
    }
  });
}

// src/app/graph-populate.class.ts
var GraphPopulateApplication = class {
  _app;
  __hooks;
  // @ts-expect-error setup in 'enableHooks'
  hooks;
  options;
  constructor(app, options) {
    this._app = app;
    this.options = options;
    const methods = ["find", "get", "create", "update", "patch", "remove"];
    const types = ["before", "after"];
    enableHooks(this, methods, types);
  }
  withAppParams(params, method, service) {
    const serviceHooks = service?.graphPopulate?.__hooks;
    if (!this.__hooks && !serviceHooks) {
      if (!params) {
        return [];
      }
      if (Array.isArray(params)) {
        return params;
      } else {
        return [params];
      }
    }
    const currentParams = [];
    const before = getHooks(this, service.graphPopulate, "before", method);
    if (before.length > 0) {
      currentParams.push(...before);
    }
    if (params) {
      if (Array.isArray(params)) {
        currentParams.push(...params);
      } else {
        currentParams.push(params);
      }
    }
    const after = getHooks(this, service.graphPopulate, "after", method, true);
    if (after.length > 0) {
      currentParams.push(...after);
    }
    return currentParams;
  }
  get allowUnnamedQueryForExternal() {
    return this.options?.allowUnnamedQueryForExternal;
  }
};

// src/app/graph-populate.service-mixin.ts
var serviceMixin = (service) => {
  if (!service.graphPopulate) {
    service.graphPopulate = {};
  }
  const methods = ["find", "get", "create", "update", "patch", "remove"];
  const types = ["before", "after"];
  enableHooks(service.graphPopulate, methods, types);
};

// src/app/graph-populate.app.ts
function initApp(options) {
  return (app) => {
    const graphPopulate2 = new GraphPopulateApplication(app, options);
    app.graphPopulate = graphPopulate2;
    app.mixins.push(serviceMixin);
  };
}
export {
  initApp as default,
  definePopulates,
  getQuery,
  graphPopulate,
  paramsForServer,
  paramsFromClient,
  populate,
  populateUtil,
  shallowPopulate
};
